@@ -22,11 +22,22 @@
 #include <stdio.h>
 #include <string.h>
 #include <assert.h>
-#include "calctool.h"
-#include <gdk/gdkkeysyms.h>
-#include "extern.h"
+
+#include "display.h"
+
+#include "mp.h"
 #include "mpmath.h"
 #include "functions.h"
+#include "ui.h"
+
+static char digits[] = "0123456789ABCDEF";
+
+static double max_fix[MAXBASES] = {
+    1.298074214e+33,    /* Binary. */
+    2.037035976e+90,    /* Octal. */
+    1.000000000e+100,   /* Decimal */
+    2.582249878e+120    /* Hexadecimal. */
+};
 
 static char *make_eng_sci(int *, int);
 
@@ -92,7 +103,7 @@ localize_number(char *dest, const char *src)
 }
 
 
-int
+static int
 char_val(char chr)
 {
     if (chr >= '0' && chr <= '9') {
@@ -111,55 +122,34 @@ void
 clear_display(int initialise)
 {
     int i;
-
+    
     v->pointed = 0;
     v->toclear = 1;
     i = 0;
     mpcim(&i, v->MPdisp_val);
     STRNCPY(v->display, make_number(v->MPdisp_val, v->base, FALSE), 
             MAXLINE - 1);
-    set_display(v->display, FALSE);
-    v->ghost_zero = 1;
+    ui_set_display(v->display, -1);
 
     if (initialise == TRUE) {
         v->show_paren = 0;
-        v->opsptr     = 0;            /* Clear parentheses stacks. */
         v->numsptr    = 0;
         v->noparens   = 0;
-        set_hyp_item(FALSE);          /* Also clears v->hyperbolic. */
-        set_inv_item(FALSE);          /* Also clears v->inverse. */
+        ui_set_hyperbolic_state(FALSE);          /* Also clears v->hyperbolic. */
+        ui_set_inverse_state(FALSE);          /* Also clears v->inverse. */
     }
 }
 
 
-/* FIXME: not a right location for this function. */
-
-struct button *
-copy_button_info(struct button *old)
-{
-    struct button *new;
-
-    int len = sizeof(struct button);
-    new = malloc(sizeof(struct button));
-    assert(new);
-    memcpy(new, old, len);
-
-    return(new);
-}
-
-
+/* TODO: perhaps this function should be renamed to reset. */
 void
 initialise()
 {
-    /* TODO: perhaps this function should be renamed to reset. */
-
     int i;
-    struct exprm_state *e;
  
     v->error         = 0;           /* Currently no display error. */
-    v->cur_op        = '?';         /* No arithmetic operator defined yet. */
-    v->old_cal_value = '?';
-    v->pending       = 0;
+    v->cur_op        = -1;         /* No arithmetic operator defined yet. */
+    v->old_cal_value = -1;
     i = 0;
     mpcim(&i, v->MPresult);         /* No previous result yet. */
     mpcim(&i, v->MPdisp_val);         
@@ -167,9 +157,7 @@ initialise()
   
     v->new_input = 1;               /* Value zero is on calculator display */
 
-    e = get_state();
-    free(e->expression);
-    e->expression = NULL;
+    exp_clear();
 }
 
 
@@ -256,10 +244,10 @@ make_fixed(int *MPnumber, char *str, int base, int cmax, int toclear)
 /* Convert MP number to character string in the given base. */
 
 char *
-make_number(int *MPnumber, int base, BOOLEAN ignoreError)
+make_number(int *MPnumber, int base, int ignoreError)
 {
     double number, val;
-
+    
 /*  NOTE: make_number can currently set v->error when converting to a double.
  *        This is to provide the same look&feel as V3 even though gcalctool
  *        now does internal arithmetic to "infinite" precision.
@@ -296,7 +284,7 @@ make_eng_sci(int *MPnumber, int base)
     int ddig;                   /* Number of digits in exponent. */
     int eng = 0;                /* Set if this is an engineering number. */
     int exp = 0;                /* Exponent */
-
+    
     if (v->dtype == ENG) {
         eng = 1;
     }
@@ -402,7 +390,7 @@ MPstr_to_num(char *str, enum base_type base, int *MPval)
     int exp      = 0;
     int exp_sign = 1;
     int negate = 0;
-    char *lnp = get_localized_numeric_point();
+    char *lnp = ui_get_localized_numeric_point();
     assert(lnp);
 
     i = 0;
@@ -468,9 +456,10 @@ MPstr_to_num(char *str, enum base_type base, int *MPval)
 /* Append the latest parenthesis char to the display item. */
 
 void
-paren_disp(int c)
+paren_disp(int key)
 {
     int i, n;
+    char *text;
 
 /*  If the character is a Delete, clear the whole line, and exit parenthesis
  *  processing.
@@ -483,14 +472,17 @@ paren_disp(int c)
  */
 
     n = strlen(v->display);
-    if (c == -1 || c == 127) {         /* Is it a Delete character? */
-        v->noparens = v->pending = v->opsptr = v->numsptr = 0;
-        v->cur_op = '?';
+    text = buttons[key].symname;
+    switch (key) {
+    case -1:
+    case KEY_CLEAR:
+        v->noparens = v->numsptr = 0;
+        v->cur_op = -1;
         i = 0;
         mpcim(&i, v->MPdisp_val);
         show_display(v->MPdisp_val);
         return;
-    } else if (c == 8) {  /* Is is a Back Space character? */
+    case KEY_BACKSPACE:
         if (!n) {
             return;
         }
@@ -500,88 +492,64 @@ paren_disp(int c)
         } else if (v->display[n-1] == '(') {
             v->noparens--;
             if (!v->noparens) {
-                v->pending = v->opsptr = v->numsptr = 0;
-                v->cur_op = '?';
+                v->numsptr = 0;
+                v->cur_op = -1;
                 show_display(v->MPdisp_val);
                 return;
             }
         } else if (v->display[n-1] == ')') v->noparens++;
         v->display[n-1] = '\0';
+        break;
 
-    } else if (c == '(') {
+    case KEY_START_BLOCK:
 
 /* If this is the first left parenthesis being displayed and there is no
  * current arithmetic operand, then the current display is initially cleared
  * to avoid the confusion of showing something like "0(".
  */
 
-        if (v->noparens == 1 && v->cur_op == '?') {
+        if (v->noparens == 1 && v->cur_op == -1) {
             n = 0;
-            v->display[n] = '\0';
+            v->display[0] = '\0';
         }
+        text = "(";
+        break;
+        
+    case KEY_END_BLOCK:
+        text = ")";
+        break;
+    }
 
-        if (n < MAXLINE-1) {
-            v->display[n]   = c;
-            v->display[n+1] = '\0';
-        }
-
-    } else {                           /* It must be an ordinary character. */
-        if (n < MAXLINE-1) {
-            v->display[n]   = c;
-            v->display[n+1] = '\0';
-        }
+    if (text) {
+        SNPRINTF(v->display+n, MAXLINE-n, "%s", text);
     }
 
     n = (n < MAX_DIGITS) ? 0 : n - MAX_DIGITS;
-    v->show_paren = 1;       /* Hack to get set_display to really display it. */
-    set_display(&v->display[n], FALSE);
+    v->show_paren = 1;       /* Hack to get ui_set_display to really display it. */
+    ui_set_display(&v->display[n], -1);
     v->show_paren = 0;
 }
 
 
 void
-process_item(struct button *button)
+process_item(struct button *button, int arg)
 {
-    int i, isvalid;
-
-    if (v->current != NULL) {
-        free(v->current);
-    }
-    v->current = copy_button_info(button);
-
-/* Reassign "extra" values. */
-
-    if (v->current->value[0] == GDK_Return) {
-        v->current->value[0] = '=';
-    }
-    if (v->current->value[0] == 'Q') {
-        v->current->value[0] = 'q';
-    }
+    v->current = button->id;
 
     if (v->error) {
-        isvalid = 0;                    /* Must press a valid key first. */
-        for (i = 0; i < MAXVKEYS; i++) {
-            if (v->current->value[0] == validkeys[i]) {
-                isvalid = 1;
-            }
-        }
-        if (v->pending == '?') {
-            isvalid = 1;
-        }
-        if (!isvalid) {
+        /* Must press a valid key first. */
+        if (v->current != KEY_CLEAR) {
             return;
         }
-        set_error_state(FALSE);
+        ui_set_error_state(FALSE);
     }
-
-    if (v->pending) {
-        (*v->pending_but->func)();
+    
+    if (v->noparens > 0) {
+        do_paren();
         return;
     }
 
-    STRCPY(v->opstr, v->op_item_text);
-
-    (*button->func)();
+    (*button->func)(arg);
 }
 
 
@@ -590,7 +558,7 @@ show_display(int *MPval)
 {
     if (!v->error) {
         STRNCPY(v->display, make_number(MPval, v->base, FALSE), MAXLINE - 1);
-        set_display(v->display, FALSE);
+        ui_set_display(v->display, -1);
     }
 }
 
@@ -598,56 +566,69 @@ show_display(int *MPval)
 /* In arithmetic precedence mode this routine should be called to redraw 
  * the display.
  */
-
 void
-refresh_display()
+refresh_display(int cursor)
 {
-    char localized[MAX_LOCALIZED];
+    int i, MP_reg[MP_SIZE];
+    char localized[MAX_LOCALIZED], *str, *ans, reg[3];
+    struct exprm_state *e;
 
     switch (v->syntax) {
-        case npa:
+        case NPA:
             show_display(v->MPdisp_val);
             break;
 
-        case exprs: {
-	    struct exprm_state *e = get_state();
-
-            if (e->expression && strlen(e->expression)) {
-                char *str = gc_strdup(e->expression);
-                char *ans = make_number(e->ans, v->base, TRUE);
-
-                localize_number(localized, ans);
-                str_replace(&str, "Ans", localized);
-
-                { /* Replace registers with values. */
-                    char reg[3];
-                    int n = '0';
-                    int i;
-
-                    for (i = 0; i < 10; i++) {
-                        char *reg_val;
-                        int MP_reg[MP_SIZE];
-
-                        snprintf(reg, 3, "R%c", n+i);
-                        do_rcl_reg(i, MP_reg);
-                        reg_val = make_number(MP_reg, v->base, FALSE);
-                        str_replace(&str, reg, reg_val);
-                    }
-                }
-                write_display(str);
-                free(str);
-                v->ghost_zero = 0;
-            } else {
-                int MP1[MP_SIZE];
-
-                do_zero(MP1);
-                show_display(MP1);
-                v->ghost_zero = 1;
+        case EXPRS:
+            e = get_state();
+            if (e->expression[0] == '\0') {
+                do_zero(MP_reg);
+                str = gc_strdup(make_number(MP_reg, v->base, FALSE));
+            } else {           
+                str = gc_strdup(e->expression);
             }
-        }    
-        break;
+            ans = make_number(e->ans, v->base, TRUE);
+
+            localize_number(localized, ans);
+            str_replace(&str, "Ans", localized);
+
+            /* Replace registers with values. */
+            for (i = 0; i < 10; i++) {
+                SNPRINTF(reg, 3, "R%d", i);
+                do_rcl_reg(i, MP_reg);
+                str_replace(&str, reg, make_number(MP_reg, v->base, FALSE));
+            }
+
+            ui_set_display(str, cursor);
+            free(str);
+            break;
 
         default:
             assert(0);
     }
 }
+
+gboolean display_is_result(void)
+{
+    struct exprm_state *e;
+
+    switch (v->syntax) {
+        case NPA:
+            if (v->old_cal_value < 0 ||
+                v->old_cal_value == KEY_CALCULATE) {
+                return TRUE;
+            }
+            break;
+
+        case EXPRS:
+            e = get_state();
+            if (strcmp(e->expression, "Ans") == 0) {
+                return TRUE;
+            }
+            break;
+        
+        default:
+            assert(FALSE);
+    }
+    
+    return FALSE;
+}
